/* 
 * makedumpfile.c
 *
 * Copyright (C) 2006  NEC Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/*
 * TODO
 * 1. get a structure size and a field offset. : Done.
 * 2. get the memory management information from the symbol "pgdat_list".
 * 3. exclude free pages.
 * 4. (i386) fill PT_LOAD headers with appropriate virtual addresses.
 */

#include "makedumpfile.h"

/*
 *  Strip line-ending linefeeds in a string.
 */
char *
strip_linefeeds(char *line)
{
	char *p;

	if (line == NULL || strlen(line) == 0)
		return(line);

	p = &LASTCHAR(line);

	while (*p == '\n') {
		*p = NULLCHAR;
		if (--p < line)
			break; 
	}

	return(line);
}

/*
 *  Strip line-beginning and line-ending whitespace and linefeeds.
 */
char *
clean_line(char *line)
{
	line = g_strchug(line);
	strip_linefeeds(line);
	line = g_strchomp(line);
	return(line);
}

/*
 *  Parse a line into tokens, populate the passed-in argv[] array, and return
 *  the count of arguments found.  This function modifies the passed-string 
 *  by inserting a NULL character at the end of each token.  Expressions 
 *  encompassed by parentheses, and strings encompassed by apostrophes, are 
 *  collected into single tokens.
 */
int
parse_line(char *str, char *argv[])
{
	int i, j;
	int string;
	int expression;

	for (i = 0; i < MAXARGS; i++)
		argv[i] = NULL;

	clean_line(str);

	if (str == NULL || strlen(str) == 0)
		return(0);

	i = j = 0;
	string = expression = FALSE;
	argv[j++] = str;

	while (TRUE) {
		if (j == MAXARGS)
			ERRMSG("Too many arguments in string!\n");

		while (str[i] != ' ' && str[i] != '\t' && str[i] != NULLCHAR) {
			i++;
		}

		switch (str[i])
		{
		case ' ':
		case '\t':
			str[i++] = NULLCHAR;
	
		    if (str[i] == '"') {    
			str[i] = ' ';
			string = TRUE;
			i++;
		    }

		    if (str[i] == '(') {     
			expression = TRUE;
		    }

		    while (str[i] == ' ' || str[i] == '\t') {
			i++;
		    }

		    if (str[i] != NULLCHAR && str[i] != '\n') {
			argv[j++] = &str[i];
			if (string) {
				string = FALSE;
				while (str[i] != '"' && str[i] != NULLCHAR)
					i++;
				if (str[i] == '"')
					str[i] = ' ';
			}
			if (expression) {
				expression = FALSE;
				while (str[i] != ')' && str[i] != NULLCHAR)
					i++;
			}
			break;
		    }
		/* else fall through */
		case '\n':
		    str[i] = NULLCHAR;
		/* keep falling... */
		case NULLCHAR:
		    argv[j] = NULLCHAR;
		    return(j);
		}
	}
}

/*
 * Convert Physical Address to File Offest.
 *  If this function returns 0x0, File Offset isn't found.
 *  The File Offset 0x0 is in the ELF header.
 *  It is not in the memory image.
 */
off_t
paddr_to_fileoffset(struct DumpInfo *info,  unsigned long paddr)
{
	int i;
	off_t offset;
	struct pt_load_segment *pls;

	for (i = offset = 0; i < info->num_load_memory; i++) {
		pls = &info->pt_load_segments[i];
		if ((paddr >= pls->phys_start)
		    && (paddr < pls->phys_end)) {
			offset = (off_t)(paddr - pls->phys_start) +
				pls->file_offset;
				break;
		}
	}
	return offset;
}

/*
 * Convert Virtual Address to File Offest.
 *  If this function returns 0x0, File Offset isn't found.
 *  The File Offset 0x0 is the ELF header.
 *  It is not in the memory image.
 */
off_t
vaddr_to_fileoffset(struct DumpInfo *info,  unsigned long vaddr)
{
	int i;
	off_t offset;
	struct pt_load_segment *pls;

	for (i = offset = 0; i < info->num_load_memory; i++) {
		pls = &info->pt_load_segments[i];
		if ((vaddr >= pls->virt_start)
		    && (vaddr < pls->virt_end)) {
			offset = (off_t)(vaddr - pls->virt_start) +
				pls->file_offset;
				break;
		}
	}
	return offset;
}

/*
 * Get the number of the page descriptors from the ELF info.
 */
unsigned int
num_page_descriptor(struct DumpInfo *info)
{
	int i;
	off_t max_paddr;
	struct pt_load_segment *pls;

	for (i = 0, max_paddr = 0; i < info->num_load_memory; i++) {
		pls = &info->pt_load_segments[i];
		if (max_paddr < pls->phys_end)
			max_paddr = pls->phys_end;
	}
	return max_paddr / info->page_size;
}

int
readmem(struct DumpInfo *info, unsigned long vaddr, void *bufptr, size_t size)
{
	off_t offset;
	const off_t failed = (off_t)-1;

	/*
	 * Convert Virtual Address to File Offset.
	 */
	if (!(offset = vaddr_to_fileoffset(info, vaddr)))
		return FALSE;

	if (lseek(info->fd_memory, offset, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		return FALSE;
	}

	if (read(info->fd_memory, bufptr, size) != size) {
		ERRMSG("Can't read the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		return FALSE;
	}

	return size;
}

void
print_usage()
{
	MSG("\n");
	MSG("Usage:\n");
	MSG("  makedumpfile [-c|-E] [-d dump_level] [-x vmlinux] [map_file] dump_mem dump_file\n");
	MSG("\n");
	MSG("Available options:\n");
	MSG("  [-c]:\n");
	MSG("      This option enables the compression function of each page.\n");
	MSG("      This is only for crash.\n");
	MSG("\n");
	MSG("  [-E]:\n");
	MSG("      Create the ELF dump file.\n");
	MSG("      You can not specify this opiton with [-c].\n");
	MSG("      This is only for gdb.\n");
	MSG("\n");
	MSG("  [-d dump_level]:\n");
	MSG("      This is specification of the skipped pages. \n");
	MSG("      The page type marked in the following table is skipped.\n");
	MSG("\n");
	MSG("      dump  |  zero   cache   cache    user\n");
	MSG("      level |  page   page    private  data\n");
	MSG("     ---------------------------------------\n");
	MSG("         0  |\n");
	MSG("         1  |  X\n");
	MSG("         2  |         X\n");
//	MSG("         3  |  X      X\n");
	MSG("         4  |         X       X\n");
//	MSG("         5  |  X      X       X\n");
//	MSG("         6  |         X       X\n");
//	MSG("         7  |  X      X       X\n");
	MSG("         8  |                          X\n");
//	MSG("         9  |  X                       X\n");
//	MSG("        10  |         X                X\n");
//	MSG("        11  |  X      X                X\n");
//	MSG("        12  |         X       X        X\n");
//	MSG("        13  |  X      X       X        X\n");
//	MSG("        14  |         X       X        X\n");
	MSG("        15  |  X      X       X        X\n");
	MSG("\n");
	MSG("  [-x vmlinux]:\n");
	MSG("      This is a pathname to a vmlinux file compiled with -g option\n");
	MSG("      for first-kernel.\n");
	MSG("      This is necessary when dump_level is 2 or more.\n");
	MSG("\n");
	MSG("  [map_file]:\n");
	MSG("      This is a pathname to a system.map file of first-kernel.\n");
	MSG("      This is necessary when dump_level is 2 or more.\n");
	MSG("\n");
	MSG("  dump_mem:\n");
	MSG("      This is a pathname to a first-kernel memory core image.\n");
	MSG("      This argument is generally /proc/vmcore.\n");
	MSG("\n");
	MSG("  dump_file:\n");
	MSG("      This is a pathname to a filename created by this command.\n");
	MSG("\n");
}

int
open_system_map(struct DumpInfo *info)
{
	FILE *map;

	if (info->name_sysmap == NULL)
		return TRUE;

	if ((map = fopen(info->name_sysmap, "r")) == NULL) {
		ERRMSG("Can't open the map file(%s). %s\n",
		    info->name_sysmap, strerror(errno));
		return FALSE;
	}

	info->map = map;
	return TRUE;
}

int
open_dump_memory(struct DumpInfo *info)
{
	int fd;

	if ((fd = open(info->name_memory, O_RDONLY)) < 0) {
		ERRMSG("Can't open the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		return FALSE;
	}
	info->fd_memory = fd;

	return TRUE;
}

int
open_dump_file(struct DumpInfo *info)
{
	int fd;

	if ((fd = open(info->name_dumpfile, O_RDWR|O_CREAT|O_EXCL,
	    S_IRUSR|S_IWUSR)) < 0) {
		ERRMSG("Can't open the dump file(%s). %s\n",
		    info->name_dumpfile, strerror(errno));
		return FALSE;
	}
	info->fd_dumpfile = fd;
	return TRUE;
}

int
open_dump_bitmap(struct DumpInfo *info)
{
	int fd;

	if ((info->name_bitmap
	    = (char *)g_malloc(sizeof(FILENAME_BITMAP))) == NULL) {
		ERRMSG("Can't allocate memory for the filename. %s\n",
		    strerror(errno));
		return FALSE;
	}
	strcpy(info->name_bitmap, FILENAME_BITMAP);
	if ((fd = open(info->name_bitmap, O_RDWR|O_CREAT,
	    S_IRUSR|S_IWUSR)) < 0) {
		ERRMSG("Can't open the dump file(%s). %s\n",
		    FILENAME_BITMAP, strerror(errno));
		return FALSE;
	}
	unlink(info->name_bitmap);
	info->fd_bitmap = fd;
	return TRUE;
}

int
dump_Elf32_pt_load(struct DumpInfo *info, Elf32_Phdr *prog, int num_load)
{
	struct pt_load_segment *pls;

	if (prog->p_type != PT_LOAD) {
		ERRMSG("%s isn't the dump memory.\n", info->name_memory);
		return FALSE;
	}

	pls = &info->pt_load_segments[num_load];
	pls->phys_start  = prog->p_paddr;
	pls->phys_end    = pls->phys_start + prog->p_filesz;
	pls->virt_start  = prog->p_vaddr;
	pls->virt_end    = pls->virt_start + prog->p_filesz;
	pls->file_offset = prog->p_offset;

	return TRUE;
}

int
dump_Elf64_pt_load(struct DumpInfo *info, Elf64_Phdr *prog, int num_load)
{
	struct pt_load_segment *pls;

	if (prog->p_type != PT_LOAD) {
		ERRMSG("%s isn't the dump memory.\n", info->name_memory);
		return FALSE;
	}

	pls = &info->pt_load_segments[num_load];
	pls->phys_start  = prog->p_paddr;
	pls->phys_end    = pls->phys_start + prog->p_filesz;
	pls->virt_start  = prog->p_vaddr;
	pls->virt_end    = pls->virt_start + prog->p_filesz;
	pls->file_offset = prog->p_offset;
	return TRUE;
}

int
get_elf_info(struct DumpInfo *info)
{
	int i;
	const off_t failed = (off_t)-1;

	Elf32_Ehdr *elf32;
	Elf64_Ehdr *elf64;
	Elf32_Phdr *load32;
	Elf64_Phdr *load64;
	char header[MIN_ELF_HEADER_SIZE];
	char *tmp_elf_header = NULL;

	/*
	 * Get minimam ELF header info.
	 */
	if (read(info->fd_memory, header, MIN_ELF_HEADER_SIZE)
	    != MIN_ELF_HEADER_SIZE) {
		ERRMSG("Can't read the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		goto out;
	}

	elf32 = (Elf32_Ehdr *)&header[0];
	elf64 = (Elf64_Ehdr *)&header[0];

	if (STRNEQ(elf32->e_ident, ELFMAG)
	    && (elf32->e_ident[EI_CLASS] == ELFCLASS32)) {
		info->flag_elf = ELF32;
		load32 = (Elf32_Phdr *)
		       &header[sizeof(Elf32_Ehdr)+sizeof(Elf32_Phdr)];
		info->num_load_memory    = elf32->e_phnum - 1;
		info->offset_load_memory = (Elf32_Off)load32->p_offset;
		info->page_size = (uint32_t)load32->p_align;
	} else if (STRNEQ(elf64->e_ident, ELFMAG)
	    && (elf64->e_ident[EI_CLASS] == ELFCLASS64)) {
		info->flag_elf = ELF64;
		load64 = (Elf64_Phdr *)
		       &header[sizeof(Elf64_Ehdr)+sizeof(Elf64_Phdr)];
		info->num_load_memory    = elf64->e_phnum - 1;
		info->offset_load_memory = (Elf64_Off)load64->p_offset;
		info->page_size = (uint64_t)load64->p_align;
	} else {
		ERRMSG("%s isn't the dump memory.\n", info->name_memory);
		goto out;
	}

	/*
	 * FIXME
	 *   If the page_size of 1st-kernel is different from the one of
	 *   capture(2nd)-kernel, the problem will happen.
	 */
	if (!info->page_size)
		info->page_size = (size_t)sysconf(_SC_PAGE_SIZE);
	if (!info->page_size) {
		ERRMSG("Can't get the size of page.\n");
		goto out;
	}
	if (!info->num_load_memory) {
		ERRMSG("Can't get the number of PT_LOAD.\n");
		goto out;
	}
	if (!info->offset_load_memory) {
		ERRMSG("Can't get the offset of page data.\n");
		goto out;
	}

	/*
	 * Get full ELF header info.
	 */
	if ((tmp_elf_header = (char *)g_malloc0(info->offset_load_memory))
	    == NULL) {
		ERRMSG("Can't allocate memory for the ELF header. %s\n",
		    strerror(errno));
		goto out;
	}
	if (lseek(info->fd_memory, 0, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		goto out;
	}
	if (read(info->fd_memory, tmp_elf_header, info->offset_load_memory)
	    != info->offset_load_memory) {
		ERRMSG("Can't read the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		goto out;
	}
	if (info->flag_elf & ELF32) {
		load32 = (Elf32_Phdr *)
		       &tmp_elf_header[sizeof(Elf32_Ehdr)+sizeof(Elf32_Phdr)];
		if ((info->pt_load_segments = (struct pt_load_segment *)
		    g_malloc0(sizeof(struct pt_load_segment) *
		    info->num_load_memory)) == NULL) {
			ERRMSG("Can't allocate memory for the PT_LOAD. %s\n",
			    strerror(errno));
			goto out;
		}
		for (i = 0; i < info->num_load_memory; i++)
			if(!dump_Elf32_pt_load(info, load32 + i, i))
				goto out;
	} else { /* ELF64 */
		load64 = (Elf64_Phdr *)
		       &tmp_elf_header[sizeof(Elf64_Ehdr)+sizeof(Elf64_Phdr)];
		if ((info->pt_load_segments = (struct pt_load_segment *)
		    g_malloc0(sizeof(struct pt_load_segment) *
		    info->num_load_memory)) == NULL) {
			ERRMSG("Can't allocate memory for the PT_LOAD. %s\n",
			    strerror(errno));
			goto out;
		}
		for (i = 0; i < info->num_load_memory; i++)
			if(!dump_Elf64_pt_load(info, load64 + i, i))
				goto out;
	}
	g_free(tmp_elf_header);
	return TRUE;
out:
	if (tmp_elf_header != NULL)
		g_free(tmp_elf_header);
	return FALSE;
}

int
get_symbol_info(struct DumpInfo *info)
{
	unsigned int max_mapnr;
	unsigned long sym_mem_map = 0, sym_num_physpages = 0;
	unsigned long sym_system_utsname = 0;
	unsigned long addr_mem_map;
	struct new_utsname system_utsname;
	char buf[BUFSIZE_FGETS], *mapitems[MAXARGS], *endp;

	if (info->name_sysmap == NULL) {
		info->max_mapnr = num_page_descriptor(info);
		info->len_bitmap = 2*divideup(info->max_mapnr, BITPERBYTE);
		return TRUE;
	}

	/*
	 * Get a symbol info from the System.map.
	 */
	while (fgets(buf, BUFSIZE_FGETS, info->map)) {
		if (parse_line(buf, mapitems) != 3)
			continue;

		if (strcmp(mapitems[2], "mem_map") == 0) {
			sym_mem_map = strtoul(mapitems[0], &endp, 16);
        		if ((!sym_mem_map || sym_mem_map == ULONG_MAX)
			    || strlen(endp) != 0) {
				ERRMSG("The map file has invalid address of %s",
				    mapitems[2]);
				return FALSE;
			}
		}
		if (strcmp(mapitems[2], "num_physpages") == 0) {
			sym_num_physpages = strtoul(mapitems[0], &endp, 16);
        		if ((!sym_num_physpages
			    || sym_num_physpages == ULONG_MAX)
			    || strlen(endp) != 0) {
				ERRMSG("The map file has invalid address of %s",
				    mapitems[2]);
				return FALSE;
			}
		}
		if (strcmp(mapitems[2], "system_utsname") == 0) {
			sym_system_utsname = strtoul(mapitems[0], &endp, 16);
        		if ((!sym_system_utsname
			    || sym_system_utsname == ULONG_MAX)
			    || strlen(endp) != 0) {
				ERRMSG("The map file has invalid address of %s",
				    mapitems[2]);
				return FALSE;
			}
		}
		if ((sym_mem_map != 0)
		    && (sym_num_physpages != 0)
		    && (sym_system_utsname != 0))
			break;
	}

	/*
	 * Get the address of the symbol "mem_map".
	 * FIXME
	 *   If this command get the feature of analysing a struct,
	 *   the info of the memory management will be gotten from "pgdat_list".
	 */
	if (!sym_mem_map) {
		ERRMSG("Can't get the symbol of mem_map.\n");
		return FALSE;
	}
	if (!readmem(info, sym_mem_map, &addr_mem_map, sizeof addr_mem_map)) {
		ERRMSG("Can't get the address of mem_map.\n");
		return FALSE;
	}

	if (!addr_mem_map) {
		ERRMSG("Can't get the address of mem_map.\n");
		return FALSE;
	}
	info->addr_mem_map = addr_mem_map;

	/*
	 * Get the number of the page descriptors.
	 */
	if (!sym_num_physpages || !readmem(info, sym_num_physpages, &max_mapnr,
	     sizeof max_mapnr) || !max_mapnr)
			max_mapnr = num_page_descriptor(info);

	if (!max_mapnr) {
		ERRMSG("Can't get the number of the page descriptors.\n");
		return FALSE;
	}
	info->max_mapnr = max_mapnr;
	info->len_bitmap = 2*divideup(info->max_mapnr, BITPERBYTE);

	/*
	 * Get the kernel version from the symbol "system_utsname".
	 */
	if (!sym_system_utsname) {
		ERRMSG("Can't get the symbol of system_utsname.\n");
		return FALSE;
	}
	if (!readmem(info, sym_system_utsname, &system_utsname, sizeof(struct new_utsname))) {
		ERRMSG("Can't get the address of system_utsname.\n");
		return FALSE;
	}

	if (!strncmp(system_utsname.release, "2.6.15", strlen("2.6.15"))) {
		info->kernel_version = VERSION_2_6_15;
	} else if (!strncmp(system_utsname.release, "2.6.16", strlen("2.6.16"))) {
		info->kernel_version = VERSION_2_6_16;
	} else if (!strncmp(system_utsname.release, "2.6.17", strlen("2.6.17"))) {
		info->kernel_version = VERSION_2_6_17;
	} else {
		ERRMSG("Can't get kernel version from system_utsname.\n");
		return FALSE;
	}

	return TRUE;
}

int
get_pagedesc_offset(struct DumpInfo *info)
{

	/*
	 * Get a offset of the page_discriptor's member.
	 */

	dwarf_info.status = 0;
	dwarf_info.struct_name = "page";
	dwarf_info.file_name = "mm.h";
	dwarf_info.struct_size = 0;
	dwarf_info.no_of_member = 3;
	dwarf_info.cur_member = DWARF_INFO_MAX;
	dwarf_info.member[0].name = "flags";
	dwarf_info.member[0].offset = 0;
	dwarf_info.member[1].name = "_count";
	dwarf_info.member[1].offset = 0;
	if (info->kernel_version == VERSION_2_6_15) { /* linux-2.6.15 */
		dwarf_info.location_member = DWARF_INFO_MAX;
		dwarf_info.member[2].name = "mapping";
		dwarf_info.member[2].offset = 0;
	} else {
		dwarf_info.location_member = 2;
		dwarf_info.member[2].name = NULL; /* has no name */
		dwarf_info.member[2].offset = 0;
	}
	if (dwarfinfo())
		return FALSE;

	info->size_pagedesc = dwarf_info.struct_size;
	info->offset_page_flags  = dwarf_info.member[0].offset;
	info->offset_page__count = dwarf_info.member[1].offset;
	if (info->kernel_version == VERSION_2_6_15) { /* linux-2.6.15 */
		/* 'mapping' found. For linux-2.6.15. */
		info->offset_page_mapping = dwarf_info.member[2].offset;
	} else {
		/* 'mapping' is in union. For linux-2.6.16 or after. */
		info->offset_page_mapping = dwarf_info.member[2].offset
		+ sizeof(unsigned long);
	}

	return TRUE;
}

int
initial(struct DumpInfo *info)
{
	if (!get_elf_info(info))
		return FALSE;

	if (!get_symbol_info(info))
		return FALSE;

	if (!get_pagedesc_offset(info))
		return FALSE;

	return TRUE;
}

static inline void
set_bitmap(char *bitmap, unsigned int pfn, int val)
{
	int byte, bit;

	byte = pfn>>3;
	bit  = pfn & 7;

	if (val)
		bitmap[byte] |= 1<<bit;
	else
		bitmap[byte] &= ~(1<<bit);
}

static inline int
is_on(char *bitmap, int i)
{
	return bitmap[i>>3] & (1 << (i & 7));
}

static inline int
is_dumpable(struct dump_bitmap *bitmap, int pfn)
{
	off_t offset;
	if (pfn == 0 || bitmap->no_block != pfn/PFN_BUFBITMAP) {
		offset = bitmap->offset + BUFSIZE_BITMAP*(pfn/PFN_BUFBITMAP);
		lseek(bitmap->fd, offset, SEEK_SET);
		read(bitmap->fd, bitmap->buf, BUFSIZE_BITMAP);
		if (pfn == 0)
			bitmap->no_block = 0;
		else
			bitmap->no_block = pfn/PFN_BUFBITMAP;
	}
	return is_on(bitmap->buf, pfn%PFN_BUFBITMAP);
}

static inline int
is_memory_hole(struct dump_bitmap *bitmap, int pfn)
{
	return !is_dumpable(bitmap, pfn);
}

static inline int
is_in_segs(struct DumpInfo *info, unsigned long long paddr)
{
	if (paddr_to_fileoffset(info, paddr))
		return TRUE;
	else
		return FALSE;
}

static inline int
is_zero_page(unsigned char *buf, size_t page_size)
{
	size_t i;

	for (i = 0; i < page_size; i++)
		if (buf[i])
			return FALSE;
	return TRUE;
}

int
read_cache(struct cache_data *cd)
{
	const off_t failed = (off_t)-1;

	if (lseek(cd->fd, cd->offset, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump file(%s). %s\n",
		    cd->file_name, strerror(errno));
		return FALSE;
	}
	if (read(cd->fd, cd->buf, cd->cache_size) != cd->cache_size) {
		ERRMSG("Can't read the dump file(%s). %s\n",
		    cd->file_name, strerror(errno));
		return FALSE;
	}
	cd->offset += cd->cache_size;
	return TRUE;
}

int
write_cache(struct cache_data *cd, void *buf, size_t size)
{
	const off_t failed = (off_t)-1;

	memcpy(cd->buf + cd->buf_size, buf, size);
	cd->buf_size += size;

	if (cd->buf_size < cd->cache_size)
		return TRUE;

	if (lseek(cd->fd, cd->offset, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump file(%s). %s\n",
		    cd->file_name, strerror(errno));
		return FALSE;
	}
	if (write(cd->fd, cd->buf, cd->cache_size) != cd->cache_size) {
		ERRMSG("Can't write the dump file(%s). %s\n",
		    cd->file_name, strerror(errno));
		return FALSE;
	}

	cd->buf_size -= cd->cache_size;
	memcpy(cd->buf, cd->buf + cd->cache_size, cd->buf_size);
	cd->offset += cd->cache_size;
	return TRUE;
}

int
write_cache_bufsz(struct cache_data *cd)
{
	const off_t failed = (off_t)-1;

	if (!cd->buf_size)
		return TRUE;

	if (lseek(cd->fd, cd->offset, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump file(%s). %s\n",
		    cd->file_name, strerror(errno));
		return FALSE;
	}
	if (write(cd->fd, cd->buf, cd->buf_size) != cd->buf_size) {
		ERRMSG("Can't write the dump file(%s). %s\n",
		    cd->file_name, strerror(errno));
		return FALSE;
	}
	cd->offset  += cd->buf_size;
	cd->buf_size = 0;
	return TRUE;
}

int
create_contig_bitmap(struct DumpInfo *info)
{
	unsigned int i, pfn, remain_size, last_pfn, contig_exclude;
	unsigned int num_load_dumpfile;
	int lastpage_mhole;
	struct cache_data bm2;
	struct dump_bitmap bitmap1, bitmap2;

	bm2.fd         = info->fd_bitmap;
	bm2.file_name  = info->name_bitmap;
	bm2.cache_size = BUFSIZE_BITMAP;
	bm2.buf_size   = 0;
	bm2.offset     = info->len_bitmap/2;
	bm2.buf        = NULL;

	bitmap1.fd        = info->fd_bitmap;
	bitmap1.file_name = info->name_bitmap;
	bitmap1.no_block  = -1;
	bitmap1.buf       = NULL;
	bitmap1.offset    = 0;

	bitmap2.fd        = info->fd_bitmap;
	bitmap2.file_name = info->name_bitmap;
	bitmap2.no_block  = -1;
	bitmap2.buf       = NULL;
	bitmap2.offset    = info->len_bitmap/2;

	if ((bm2.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for 2nd-bitmap buffer. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((bitmap1.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for the 1st bitmap. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((bitmap2.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for the 2nd bitmap. %s\n",
		    strerror(errno));
		goto out;
	}
	num_load_dumpfile = info->num_load_memory;

	for (pfn = 0, last_pfn = 0,
	    contig_exclude = 0; pfn < info->max_mapnr; pfn++) {
		if (is_memory_hole(&bitmap1, pfn)) {
			for (i = 0; i <= contig_exclude; i++) {
				if ((last_pfn+i) != 0
				    && (last_pfn+i)%PFN_BUFBITMAP == 0) {
					bm2.buf_size = BUFSIZE_BITMAP;
					if (!write_cache_bufsz(&bm2))
						goto out;
				}
				set_bitmap(bm2.buf,
				    (last_pfn+i)%PFN_BUFBITMAP, 0);
			}
			contig_exclude = 0;
			last_pfn = pfn+1;
			lastpage_mhole = 1;
			continue;
		}
		lastpage_mhole = 0;
		if (!is_dumpable(&bitmap2, pfn)) {
			contig_exclude++;
			continue;
		}
		for (i = 0; i <= contig_exclude; i++) {
			if ((last_pfn+i) != 0
			    && (last_pfn+i)%PFN_BUFBITMAP == 0) {
				bm2.buf_size = BUFSIZE_BITMAP;
				if (!write_cache_bufsz(&bm2))
					goto out;
			}
			if (contig_exclude < PFN_EXCLUDED)
				set_bitmap(bm2.buf,
				    (last_pfn+i)%PFN_BUFBITMAP, 1);
			else if (i == contig_exclude)
				set_bitmap(bm2.buf,
				    (last_pfn+i)%PFN_BUFBITMAP, 1);
			else
				set_bitmap(bm2.buf,
				    (last_pfn+i)%PFN_BUFBITMAP, 0);
		}
		if (contig_exclude >= PFN_EXCLUDED)
			num_load_dumpfile++;
		contig_exclude = 0;
		last_pfn = pfn+1;
	}
	if (contig_exclude)
		num_load_dumpfile++;

	for (i = 0; i <= contig_exclude; i++) {
		if ((last_pfn+i)%PFN_BUFBITMAP == 0) {
			bm2.buf_size = BUFSIZE_BITMAP;
			if (!write_cache_bufsz(&bm2))
				goto out;
		}
		set_bitmap(bm2.buf, (last_pfn+i)%PFN_BUFBITMAP, 0);
	}
	remain_size = info->len_bitmap - bm2.offset;
	bm2.buf_size = remain_size;
	if (!write_cache_bufsz(&bm2))
		goto out;

	info->num_load_dumpfile = num_load_dumpfile;

	g_free(bm2.buf);
	g_free(bitmap1.buf);
	g_free(bitmap2.buf);
	return TRUE;
out:
	if (bm2.buf != NULL)
		g_free(bm2.buf);
	if (bitmap1.buf != NULL)
		g_free(bitmap1.buf);
	if (bitmap2.buf != NULL)
		g_free(bitmap2.buf);
	return FALSE;
}

int
create_dump_bitmap(struct DumpInfo *info)
{
	int val;
	unsigned int i, pfn, remain_size;
	unsigned long addr_mem_map, paddr;
	unsigned char *page_cache = NULL, *buf = NULL, *pcache;
	unsigned int *_count;
	unsigned long *flags, *mapping;
	struct cache_data bm1, bm2;
	off_t offset_page;
	const off_t failed = (off_t)-1;

	addr_mem_map = info->addr_mem_map;
	offset_page  = info->offset_load_memory;

	bm1.fd         = info->fd_bitmap;
	bm1.file_name  = info->name_bitmap;
	bm1.cache_size = BUFSIZE_BITMAP;
	bm1.buf_size   = 0;
	bm1.offset     = 0;
	bm1.buf        = NULL;

	bm2.fd         = info->fd_bitmap;
	bm2.file_name  = info->name_bitmap;
	bm2.cache_size = BUFSIZE_BITMAP;
	bm2.buf_size   = 0;
	bm2.offset     = info->len_bitmap/2;
	bm2.buf        = NULL;
	if ((bm1.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for 1st-bitmap buffer. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((bm2.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for 2nd-bitmap buffer. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((page_cache = g_malloc(info->size_pagedesc*PGMM_CACHED)) == NULL) {
		ERRMSG("Can't allocate memory for the pagedesc cache. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((buf = g_malloc(info->page_size)) == NULL) {
		ERRMSG("Can't allocate memory for the page. %s\n",
		    strerror(errno));
		goto out;
	}

	for (pfn = 0, paddr = 0; pfn < info->max_mapnr;
	    pfn++, addr_mem_map += info->size_pagedesc,
	    paddr += info->page_size) {

		if ((pfn != 0) && (pfn%PFN_BUFBITMAP) == 0) {
			/*
			 * Write the 1st-bitmap and the 2nd-bitmap.
			 */
			bm1.buf_size = BUFSIZE_BITMAP;
			bm2.buf_size = BUFSIZE_BITMAP;
			if (!write_cache_bufsz(&bm1))
				goto out;
			if (!write_cache_bufsz(&bm2))
				goto out;

			/*
			 * Clear the remainder of the bitmap.
			 */
			if ((info->max_mapnr - pfn) <= PFN_BUFBITMAP) {
				for (i = 0; i < PFN_BUFBITMAP; i++) {
					set_bitmap(bm1.buf, i, 0);
					set_bitmap(bm2.buf, i, 0);
				}
			}
		}
		/*
		 * val  1: dump Page
		 *      0: not dump Page
		 */
		val = 1;

		/*
		 * Exclude the memory hole.
		 */
		if (!is_in_segs(info, paddr))
			val = 0;

		/*
		 * Set the 1st-bitmap.
		 *  val  1: not memory hole
		 *       0: memory hole
		 */
		set_bitmap(bm1.buf, pfn%PFN_BUFBITMAP, val);

		/*
		 * Exclude the page filled with zero in case of creating
		 * the elf dumpfile.
		 */
		if (info->flag_elf_dumpfile
		    && (val != 0)
		    && (info->dump_level & DL_EXCLUDE_ZERO)) {
			if (lseek(info->fd_memory, offset_page, SEEK_SET)
			    == failed) {
				ERRMSG("Can't seek the dump memory(%s). %s\n",
				    info->name_memory, strerror(errno));
				goto out;
			}
			if (read(info->fd_memory, buf, info->page_size)
			    != info->page_size) {
				ERRMSG("Can't read the dump memory(%s). %s\n",
				    info->name_memory, strerror(errno));
				goto out;
			}
			offset_page += info->page_size;
			if (is_zero_page(buf, info->page_size))
				val = 0;
		}
		if (info->dump_level <= DL_EXCLUDE_ZERO) {
			set_bitmap(bm2.buf, pfn%PFN_BUFBITMAP, val);
			continue;
		}

		if ((pfn % PGMM_CACHED) == 0) {
			if (!readmem(info, addr_mem_map, page_cache,
			    info->size_pagedesc * PGMM_CACHED))
				goto out;
		}
		pcache = page_cache + ((pfn%PGMM_CACHED) * info->size_pagedesc);
		flags = (unsigned long *)(pcache + info->offset_page_flags);
		_count = (unsigned int *)(pcache + info->offset_page__count);
		mapping = (unsigned long *)(pcache + info->offset_page_mapping);

		/*
		 * Exclude the cache page without the private page.
		 */
		if ((info->dump_level & DL_EXCLUDE_CACHE)
		    && (isLRU(*flags) || isSwapCache(*flags))
		    && !isPrivate(*flags) && !isAnon(*mapping))
			val = 0;

		/*
		 * Exclude the cache page with the private page.
		 */
		else if ((info->dump_level & DL_EXCLUDE_CACHE_PRI)
		    && (isLRU(*flags) || isSwapCache(*flags))
		    && !isAnon(*mapping))
			val = 0;

		/*
		 * Exclude the data page of the user process.
		 */
		else if ((info->dump_level & DL_EXCLUDE_USER_DATA)
		    && isAnon(*mapping))
			val = 0;

		/*
		 * Set the 2nd-bitmap.
		 *  val  1: dump page
		 *       0: not dump page(the memory hole, or the excluded page)
		 */
		set_bitmap(bm2.buf, pfn%PFN_BUFBITMAP, val);
	}

	/*
	 * Write the remainder of the bitmap.
	 */
	remain_size = info->len_bitmap/2 - bm1.offset;
	bm1.buf_size = remain_size;
	bm2.buf_size = remain_size;
	if (!write_cache_bufsz(&bm1))
		goto out;
	if (!write_cache_bufsz(&bm2))
		goto out;

	if (info->flag_elf_dumpfile && !create_contig_bitmap(info))
		goto out;

	g_free(page_cache);
	g_free(buf);
	g_free(bm1.buf);
	g_free(bm2.buf);
	return TRUE;
out:
	if (page_cache != NULL)
		g_free(page_cache);
	if (buf != NULL)
		g_free(buf);
	if (bm1.buf != NULL)
		g_free(bm1.buf);
	if (bm2.buf != NULL)
		g_free(bm2.buf);
	return FALSE;
}

int
write_elf_header(struct DumpInfo *info)
{
	int i, lastpage_dumpable;
	size_t size_hdr_memory, size_Ehdr, size_Phdr, size_note;
	unsigned int pfn, pfn_start, pfn_end, num_file, num_mem;
	loff_t offset_seg, offset_note_memory, offset_note_dumpfile;
	unsigned long long  vaddr_seg, paddr_seg;
	unsigned char *header_memory = NULL;
	Elf32_Ehdr *elf32;
	Elf64_Ehdr *elf64;
	Elf32_Phdr *note32;
	Elf64_Phdr *note64;
	Elf32_Phdr *load32 = NULL;
	Elf64_Phdr *load64 = NULL;
	char *buf = NULL;
	struct pt_load_segment *pls;
	struct dump_bitmap bitmap2;
	const off_t failed = (off_t)-1;

	size_hdr_memory = info->offset_load_memory;

	bitmap2.fd        = info->fd_bitmap;
	bitmap2.file_name = info->name_bitmap;
	bitmap2.no_block  = -1;
	bitmap2.buf       = NULL;
	bitmap2.offset    = info->len_bitmap/2;

	if ((bitmap2.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for the 2nd bitmap. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((header_memory = g_malloc0(size_hdr_memory)) == NULL) {
		ERRMSG("Can't allocate memory for the ELF header. %s\n",
		    strerror(errno));
		goto out;
	}
	if (lseek(info->fd_memory, 0, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		goto out;
	}
	if (read(info->fd_memory, header_memory, size_hdr_memory)
	    != size_hdr_memory) {
		ERRMSG("Can't read the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		goto out;
	}

	/*
	 * ELF header & PT_NOTE header
	 */
	if (lseek(info->fd_dumpfile, 0, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump file(%s). %s\n",
		    info->name_dumpfile, strerror(errno));
		goto out;
	}
	if (info->flag_elf & ELF32) {
		size_Ehdr = sizeof(Elf32_Ehdr);
		size_Phdr = sizeof(Elf32_Phdr);
		elf32  = (Elf32_Ehdr *)header_memory;
		elf32->e_phnum = 1 + info->num_load_dumpfile;
		if (write(info->fd_dumpfile, elf32, size_Ehdr) != size_Ehdr) {
			ERRMSG("Can't write the dump file(%s). %s\n",
			    info->name_dumpfile, strerror(errno));
			goto out;
		}
		note32 = (Elf32_Phdr *)(header_memory + size_Ehdr);
		size_note = note32->p_filesz;
		offset_note_memory   = note32->p_offset;
		offset_note_dumpfile = size_Ehdr + size_Phdr * elf32->e_phnum;
		note32->p_offset     = offset_note_dumpfile;
		if (write(info->fd_dumpfile, note32, size_Phdr) != size_Phdr) {
			ERRMSG("Can't write the dump file(%s). %s\n",
			    info->name_dumpfile, strerror(errno));
			goto out;
		}
	} else {
		size_Ehdr = sizeof(Elf64_Ehdr);
		size_Phdr = sizeof(Elf64_Phdr);
		elf64  = (Elf64_Ehdr *)header_memory;
		elf64->e_phnum = 1 + info->num_load_dumpfile;
		if (write(info->fd_dumpfile, elf64, size_Ehdr) != size_Ehdr) {
			ERRMSG("Can't write the dump file(%s). %s\n",
			    info->name_dumpfile, strerror(errno));
			goto out;
		}
		note64 = (Elf64_Phdr *)(header_memory + size_Ehdr);
		size_note = note64->p_filesz;
		offset_note_memory   = note64->p_offset;
		offset_note_dumpfile = size_Ehdr + size_Phdr * elf64->e_phnum;
		note64->p_offset     = offset_note_dumpfile;
		if (write(info->fd_dumpfile, note64, size_Phdr) != size_Phdr) {
			ERRMSG("Can't write the dump file(%s). %s\n",
			    info->name_dumpfile, strerror(errno));
			goto out;
		}
	}
	info->offset_load_dumpfile = offset_note_dumpfile + size_note;

	/*
	 * PT_LOAD header
	 */
	if (info->flag_elf & ELF32) {
		if ((load32 = g_malloc0(size_Phdr)) == NULL) {
			ERRMSG("Can't allocate memory for PT_LOAD header. %s\n",
			    strerror(errno));
			goto out;
		}
		load32->p_type   = PT_LOAD;
		load32->p_flags  = 0;
		load32->p_offset = 0;
		load32->p_vaddr  = 0;
		load32->p_paddr  = 0;
		load32->p_filesz = 0;
		load32->p_memsz  = 0;
		load32->p_align  = 0;
	} else {
		if ((load64 = g_malloc0(size_Phdr)) == NULL) {
			ERRMSG("Can't allocate memory for PT_LOAD header. %s\n",
			    strerror(errno));
			goto out;
		}
		load64->p_type   = PT_LOAD;
		load64->p_flags  = 0;
		load64->p_offset = 0;
		load64->p_vaddr  = 0;
		load64->p_paddr  = 0;
		load64->p_filesz = 0;
		load64->p_memsz  = 0;
	}
	offset_seg = info->offset_load_dumpfile;

	for (i = 0, num_mem = 0, num_file = 0;
	    i < info->num_load_memory; i++) {
		pls = &info->pt_load_segments[i];
		paddr_seg = pls->phys_start;
		vaddr_seg = pls->virt_start;
		if (info->flag_elf & ELF32) {
			load32->p_vaddr  = vaddr_seg;
			load32->p_paddr  = paddr_seg;
			load32->p_offset = offset_seg;
		} else {
			load64->p_vaddr  = vaddr_seg;
			load64->p_paddr  = paddr_seg;
			load64->p_offset = offset_seg;
		}
		if (pls->phys_start == 0)
			pfn_start = 0;
		else
			pfn_start = pls->phys_start/info->page_size;

		if (pls->phys_end == 0)
			pfn_end = 0;
		else
			pfn_end = pls->phys_end/info->page_size;

		if (is_dumpable(&bitmap2, pfn_start)) {
			lastpage_dumpable = 1;
			num_mem  = 1;
			num_file = 1;
		} else {
			lastpage_dumpable = 0;
			num_mem  = 1;
			num_file = 0;
		}
		for (pfn = pfn_start + 1; pfn < pfn_end; pfn++) {
			if (!is_dumpable(&bitmap2, pfn)) {
				num_mem++;
				lastpage_dumpable = 0;
				continue;
			}
			if (lastpage_dumpable) {
				num_mem++;
				num_file++;
				continue;
			}
			/*
			 * Create new PT_LOAD segment.
			 */
			if (info->flag_elf & ELF32) {
				load32->p_memsz  = info->page_size*num_mem;
				load32->p_filesz = info->page_size*num_file;
				if (write(info->fd_dumpfile, load32, size_Phdr)
				    != size_Phdr) {
					ERRMSG("Can't write the dump file(%s). %s\n",
					    info->name_dumpfile, strerror(errno));
					goto out;
				}
				offset_seg += load32->p_filesz;
				if (load32->p_paddr < MAXMEM)
					load32->p_vaddr += load32->p_memsz;
				load32->p_paddr += load32->p_memsz;
				load32->p_offset = offset_seg;
			} else {
				load64->p_memsz  = info->page_size*num_mem;
				load64->p_filesz = info->page_size*num_file;
				if (write(info->fd_dumpfile, load64, size_Phdr)
				    != size_Phdr) {
					ERRMSG("Can't write the dump file(%s). %s\n",
					    info->name_dumpfile, strerror(errno));
					goto out;
				}
				offset_seg += load64->p_filesz;
				if (load64->p_paddr < MAXMEM)
					load64->p_vaddr += load64->p_memsz;
				load64->p_paddr += load64->p_memsz;
				load64->p_offset = offset_seg;
			}
			num_mem  = 1;
			num_file = 1;
			lastpage_dumpable = 1;
		}
		if (info->flag_elf & ELF32) {
			load32->p_memsz  = info->page_size*num_mem;
			load32->p_filesz = info->page_size*num_file;
			if (write(info->fd_dumpfile, load32, size_Phdr)
			    != size_Phdr) {
				ERRMSG("Can't write the dump file(%s). %s\n",
				    info->name_dumpfile, strerror(errno));
				goto out;
			}
			offset_seg += load32->p_filesz;
		} else {
			load64->p_memsz  = info->page_size*num_mem;
			load64->p_filesz = info->page_size*num_file;
			if (write(info->fd_dumpfile, load64, size_Phdr)
			    != size_Phdr) {
				ERRMSG("Can't write the dump file(%s). %s\n",
				    info->name_dumpfile, strerror(errno));
				goto out;
			}
			offset_seg += load64->p_filesz;
		}
	}
	/*
	 * Write PT_NOTE segment.
	 */
	if ((buf = g_malloc0(size_note)) == NULL) {
		ERRMSG("Can't allocate memory for the ELF header. %s\n",
		    strerror(errno));
		goto out;
	}
	if (lseek(info->fd_memory, offset_note_memory, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		goto out;
	}
	if (lseek(info->fd_dumpfile, offset_note_dumpfile, SEEK_SET)
	    == failed) {
		ERRMSG("Can't seek the dump file(%s). %s\n",
		    info->name_dumpfile, strerror(errno));
		goto out;
	}
	if (read(info->fd_memory, buf, size_note) != size_note) {
		ERRMSG("Can't read the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		goto out;
	}
	if (write(info->fd_dumpfile, buf, size_note) != size_note) {
		ERRMSG("Can't write the dump file(%s). %s\n",
		    info->name_dumpfile, strerror(errno));
		goto out;
	}

	g_free(bitmap2.buf);
	g_free(header_memory);
	if (info->flag_elf & ELF32)
		g_free(load32);
	else
		g_free(load64);
	g_free(buf);
	return TRUE;
out:
	if (bitmap2.buf != NULL)
		g_free(bitmap2.buf);
	if (header_memory != NULL)
		g_free(header_memory);
	if (load32 != NULL)
		g_free(load32);
	if (load64 != NULL)
		g_free(load64);
	if (buf != NULL)
		g_free(buf);
	return FALSE;
}

int
write_diskdump_header(struct DumpInfo *info)
{
	size_t size;
	struct disk_dump_header *dh = info->dump_header;
	const off_t failed = (off_t)-1;

	strcpy(dh->signature, DISK_DUMP_SIGNATURE);
	dh->block_size   = info->page_size;
	dh->sub_hdr_size = 1;
	dh->max_mapnr    = info->max_mapnr;
	dh->nr_cpus      = 1;
	dh->bitmap_blocks
	    = divideup(info->len_bitmap, dh->block_size);

	if (lseek(info->fd_dumpfile, 0, SEEK_SET) == failed) {
		ERRMSG("Can't seek the dump file(%s). %s\n",
		    info->name_dumpfile, strerror(errno));
		return FALSE;
	}
	size = sizeof(struct disk_dump_header);
	if (write(info->fd_dumpfile, dh, size) != size) {
		ERRMSG("Can't write the dump file(%s). %s\n",
		    info->name_dumpfile, strerror(errno));
		return FALSE;
	}
	info->offset_bitmap1
	    = (1 + dh->sub_hdr_size) * dh->block_size;

	return TRUE;
}

int
write_dump_header(struct DumpInfo *info)
{
	if (info->flag_elf_dumpfile) {
		if (!write_elf_header(info))
			return FALSE;
	} else {
		if (!write_diskdump_header(info))
			return FALSE;
	}
	return TRUE;
}

void
print_progress(int current, int end)
{
	int progress;
	time_t tm;
	static time_t last_time = 0;

	if (current < end) {
		tm = time(NULL);
		if (tm - last_time < 1)
			return;
		last_time = tm;
		progress = current * 100 / end;
	} else
		progress = 100;

	ERRMSG("\r");
	ERRMSG("[%3d %%]", progress);
}

int
write_pages(struct DumpInfo *info)
{
	unsigned int pfn, num_dump, per = info->max_mapnr/100;
	unsigned int flag_change_bitmap = 0;
	unsigned long size_out;
	struct page_desc pd;
	off_t offset_data;
	struct disk_dump_header *dh = info->dump_header;
	unsigned char *buf = NULL, *buf_out = NULL;
	struct cache_data bm2, pdesc, pdata;
	struct dump_bitmap bitmap1, bitmap2;
	const off_t failed = (off_t)-1;

	bm2.fd         = info->fd_bitmap;
	bm2.file_name  = info->name_bitmap;
	bm2.cache_size = BUFSIZE_BITMAP;
	bm2.buf_size   = 0;
	bm2.offset     = info->len_bitmap/2;
	bm2.buf        = NULL;

	pdesc.fd         = info->fd_dumpfile;
	pdesc.file_name  = info->name_dumpfile;
	pdesc.cache_size = info->page_size<<info->block_order;
	pdesc.buf_size   = 0;
	pdesc.buf        = NULL;

	pdata.fd         = info->fd_dumpfile;
	pdata.file_name  = info->name_dumpfile;
	pdata.cache_size = info->page_size<<info->block_order;
	pdata.buf_size   = 0;
	pdata.buf        = NULL;

	bitmap1.fd        = info->fd_bitmap;
	bitmap1.file_name = info->name_bitmap;
	bitmap1.no_block  = -1;
	bitmap1.buf       = NULL;
	bitmap1.offset    = 0;

	bitmap2.fd        = info->fd_bitmap;
	bitmap2.file_name = info->name_bitmap;
	bitmap2.no_block  = -1;
	bitmap2.buf       = NULL;
	bitmap2.offset    = info->len_bitmap/2;

	if ((buf = g_malloc(info->page_size)) == NULL) {
		ERRMSG("Can't allocate memory for the page. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((buf_out = g_malloc(info->page_size)) == NULL) {
		ERRMSG("Can't allocate memory for the compression buffer. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((bm2.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for 2nd-bitmap buffer. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((pdesc.buf = g_malloc(pdesc.cache_size + info->page_size))
	    == NULL) {
		ERRMSG("Can't allocate memory for the page desc buffer. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((pdata.buf = g_malloc(pdata.cache_size + info->page_size))
	    == NULL) {
		ERRMSG("Can't allocate memory for the page data buffer. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((bitmap1.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for the 1st bitmap. %s\n",
		    strerror(errno));
		goto out;
	}
	if ((bitmap2.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for the 2nd bitmap. %s\n",
		    strerror(errno));
		goto out;
	}

	if (info->flag_elf_dumpfile) {
		pdata.offset = info->offset_load_dumpfile;
	} else {
		/*
		 * Calculate the offset of the page data.
		 */
		for (pfn = 0, num_dump = 0; pfn < dh->max_mapnr; pfn++) {
			if (is_dumpable(&bitmap2, pfn))
				num_dump++;
		}
		pdesc.offset
		    = (1 + dh->sub_hdr_size + dh->bitmap_blocks)*dh->block_size;
		pdata.offset = pdesc.offset + sizeof(page_desc_t) * num_dump;
		offset_data  = pdata.offset;
	}
	/*
	 * Set a fileoffset of Physical Address 0x0.
	 */
	if (lseek(info->fd_memory, info->offset_load_memory, SEEK_SET)
	    == failed) {
		ERRMSG("Can't seek the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
		goto out;
	}

	for (pfn = 0; pfn < info->max_mapnr; pfn++) {

		if ((pfn % per) == 0)
			print_progress(pfn, info->max_mapnr);

		if ((pfn % PFN_BUFBITMAP) == 0) {
			if (flag_change_bitmap) {
				bm2.buf_size = BUFSIZE_BITMAP;
				bm2.offset  -= BUFSIZE_BITMAP;
				if (!write_cache_bufsz(&bm2))
					goto out;
			}
			if (!read_cache(&bm2))
				goto out;
			flag_change_bitmap = 0;
		}

		/*
		 * Check the memory hole.
		 */
		if (is_memory_hole(&bitmap1, pfn))
			continue;
		/*
		 * Check the excluded page.
		 */
		if (!is_dumpable(&bitmap2, pfn)) {
			if (lseek(info->fd_memory, info->page_size, SEEK_CUR)
			    == failed) {
				ERRMSG("Can't seek the dump memory(%s). %s\n",
				    info->name_memory, strerror(errno));
				goto out;
			}
			continue;
		}

		if (read(info->fd_memory, buf, info->page_size)
		    != info->page_size) {
			ERRMSG("Can't read the dump memory(%s). %s\n",
			    info->name_memory, strerror(errno));
			goto out;
		}

		if (info->flag_elf_dumpfile) {
			pd.size = info->page_size;
		} else {
			/*
			 * Exclude the page filled with zeros.
			 * In case of the elf dumpfile, the zero page had been
			 * checked.
			 */
			if ((info->dump_level & DL_EXCLUDE_ZERO)
			    && is_zero_page(buf, info->page_size)) {
				set_bitmap(bm2.buf, pfn%PFN_BUFBITMAP, 0);
				flag_change_bitmap = 1;
				continue;
			}
			/*
			 * Compress the page data.
			 */
			size_out = info->page_size;
			if (info->flag_compress
			    && (compress2(buf_out, &size_out, buf,
			    info->page_size, Z_BEST_SPEED) == Z_OK)
			    && (size_out < info->page_size)) {
				pd.flags = 1;
				pd.size  = size_out;
				memcpy(buf, buf_out, pd.size);
			} else {
				pd.flags = 0;
				pd.size  = info->page_size;
			}
			pd.page_flags = 0;
			pd.offset     = offset_data;
			offset_data  += pd.size;

			/*
			 * Write the page header.
			 */
			if (!write_cache(&pdesc, &pd, sizeof(page_desc_t)))
				goto out;
		}
		/*
		 * Write the page data.
		 */
		if (!write_cache(&pdata, buf, pd.size))
			goto out;
	}

	/*
	 * Write the remainder.
	 */
	if (!write_cache_bufsz(&pdata))
		goto out;

	if (!info->flag_elf_dumpfile) {
		if (!write_cache_bufsz(&pdesc))
			goto out;
		if (flag_change_bitmap) {
			bm2.buf_size = BUFSIZE_BITMAP;
			bm2.offset  -= BUFSIZE_BITMAP;
			if (!write_cache_bufsz(&bm2))
				goto out;
		}
	}
	/*
	 * Print the progress of the end.
	 */
	print_progress(info->max_mapnr, info->max_mapnr);

	g_free(buf);
	g_free(buf_out);
	g_free(bm2.buf);
	g_free(pdesc.buf);
	g_free(pdata.buf);
	g_free(bitmap1.buf);
	g_free(bitmap2.buf);
	return TRUE;
out:
	if (buf != NULL)
		g_free(buf);
	if (buf_out != NULL)
		g_free(buf_out);
	if (bm2.buf != NULL)
		g_free(bm2.buf);
	if (pdesc.buf != NULL)
		g_free(pdesc.buf);
	if (pdata.buf != NULL)
		g_free(pdata.buf);
	if (bitmap1.buf != NULL)
		g_free(bitmap1.buf);
	if (bitmap2.buf != NULL)
		g_free(bitmap2.buf);
	return FALSE;
}

int write_dump_bitmap(struct DumpInfo *info)
{
	struct cache_data bm;
	unsigned long buf_size;
	const off_t failed = (off_t)-1;

	if (info->flag_elf_dumpfile)
		return TRUE;

	bm.fd        = info->fd_bitmap;
	bm.file_name = info->name_bitmap;
	bm.offset    = 0;
	bm.buf       = NULL;

	if ((bm.buf = g_malloc0(BUFSIZE_BITMAP)) == NULL) {
		ERRMSG("Can't allocate memory for dump bitmap buffer. %s\n",
		    strerror(errno));
		goto out;
	}
	if (lseek(info->fd_dumpfile, info->offset_bitmap1, SEEK_SET)
	    == failed) {
		ERRMSG("Can't seek the dump file(%s). %s\n",
		    info->name_dumpfile, strerror(errno));
		goto out;
	}
	buf_size = info->len_bitmap;

	while (buf_size > 0) {
		if (buf_size >= BUFSIZE_BITMAP)
			bm.cache_size = BUFSIZE_BITMAP;
		else
			bm.cache_size = buf_size;

		if(!read_cache(&bm))
			goto out;

		if (write(info->fd_dumpfile, bm.buf, bm.cache_size)
		    != bm.cache_size) {
			ERRMSG("Can't write the dump file(%s). %s\n",
			    info->name_dumpfile, strerror(errno));
			goto out;
		}
		buf_size -= BUFSIZE_BITMAP;
	}
	g_free(bm.buf);
	return TRUE;
out:
	if (bm.buf != NULL)
		g_free(bm.buf);
	return FALSE;
}

void
close_system_map(struct DumpInfo *info)
{
	if (info->name_sysmap == NULL)
		return;

	if (fclose(info->map) < 0)
		ERRMSG("Can't close the map file(%s). %s\n",
		    info->name_sysmap, strerror(errno));
}

void
close_dump_memory(struct DumpInfo *info)
{
	if (close(info->fd_memory) < 0)
		ERRMSG("Can't close the dump memory(%s). %s\n",
		    info->name_memory, strerror(errno));
}

void
close_dump_file(struct DumpInfo *info)
{
	if (close(info->fd_dumpfile) < 0)
		ERRMSG("Can't close the dump file(%s). %s\n",
		    info->name_dumpfile, strerror(errno));
}

void
close_dump_bitmap(struct DumpInfo *info)
{
	if (close(info->fd_bitmap) < 0)
		ERRMSG("Can't close the bitmap file(%s). %s\n",
		    info->name_bitmap, strerror(errno));
	g_free(info->name_bitmap);
}

void
close_kernel_file(int fd)
{
	if (close(fd) < 0)
		ERRMSG("Can't close vmlinux file(%s). %s\n",
			dwarf_info.vmlinux_name, strerror(errno));
}

int
main(int argc, char *argv[])
{
	int opt;
	struct DumpInfo *info = NULL;

	if ((info = g_malloc0(sizeof(struct DumpInfo))) == NULL) {
		ERRMSG("Can't allocate memory for the pagedesc cache. %s.\n",
		    strerror(errno));
		goto out;
	}
	if ((info->dump_header = g_malloc0(sizeof(struct disk_dump_header)))
	    == NULL) {
		ERRMSG("Can't allocate memory for the dump header. %s\n",
		    strerror(errno));
		goto out;
	}

	while ((opt = getopt(argc, argv, "b:cd:Ex:")) != -1) {
		switch (opt) {
		case 'b':
			info->block_order = atoi(optarg);
			break;
		case 'c':
			info->flag_compress = 1;
			break;
		case 'd':
			info->dump_level = atoi(optarg);
			break;
		case 'E':
			info->flag_elf_dumpfile = 1;
			break;
		case 'x':
			dwarf_info.vmlinux_name = optarg;
			break;
		case '?':
			ERRMSG("Commandline parameter is invalid.\n");
			print_usage();
			goto out;
		}
	}
	if ((info->dump_level < MIN_DUMP_LEVEL)
	    || (MAX_DUMP_LEVEL < info->dump_level)) {
		ERRMSG("Dump_level is invalid.\n");
		print_usage();
		goto out;
	}
	if ((optind + 3 < argc) 
	    || ((info->dump_level <= 1) && (argc  < optind + 2))
	    || ((info->dump_level >= 2) && (argc != optind + 3))
	    || ((info->flag_compress == 1) && (info->flag_elf_dumpfile == 1))) {
		ERRMSG("Commandline parameter is invalid.\n");
		print_usage();
		goto out;
	}
	if ((info->dump_level <= 1) && (argc == optind + 2)) {
		info->name_sysmap   = NULL;
		info->name_memory   = argv[optind];
		info->name_dumpfile = argv[optind+1];
	} else {
		info->name_sysmap   = argv[optind];
		info->name_memory   = argv[optind+1];
		info->name_dumpfile = argv[optind+2];
	}

	if (!open_system_map(info))
		goto out;

	if (!open_dump_memory(info))
		goto out;

	if (!open_dump_file(info))
		goto out;

	if (!open_dump_bitmap(info))
		goto out;

	dwarf_info.vmlinux_fd = open_kernel_file(dwarf_info.vmlinux_name);
	if (dwarf_info.vmlinux_fd == -1)
		goto out;

	if (!initial(info))
		goto out;

	if (!create_dump_bitmap(info))
		goto out;

	if (!write_dump_header(info))
		goto out;

	if (!write_pages(info))
		goto out;

	if (!write_dump_bitmap(info))
		goto out;

	close_system_map(info);

	close_dump_memory(info);

	close_dump_file(info);

	close_dump_bitmap(info);

	close_kernel_file(dwarf_info.vmlinux_fd);

	MSG("\n");
	MSG("The dumpfile is saved to %s.\n", info->name_dumpfile);
	MSG("makedumpfile Completed.\n");
	g_free(info->pt_load_segments);
	g_free(info->dump_header);
	g_free(info);
	return COMPLETED;

out:
	ERRMSG("\n");
	ERRMSG("makedumpfile Failed.\n");
	if (info->map)
		fclose(info->map);
	if (info->fd_memory)
		close(info->fd_memory);
	if (info->fd_dumpfile)
		close(info->fd_dumpfile);
	if (info->fd_bitmap)
		close(info->fd_bitmap);
	if (info->pt_load_segments != NULL)
		g_free(info->pt_load_segments);
	if (info->dump_header != NULL)
		g_free(info->dump_header);
	if (info != NULL)
		g_free(info);
	return FAILED;
}
